library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.dcc_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity ddr_rw_ctrl is
    port (
-- reset input, should be connected to the inverted LOCK output of dll
      reset: in STD_LOGIC;
-- system clock input
      clk: in STD_LOGIC;
      mclk0: in STD_LOGIC;
-- monitored event data
-- bit 33 and 32 used to indicate header and trailer of the block
      mon_data: in STD_LOGIC_VECTOR (33 downto 0);
      mon_re: out STD_LOGIC;
      mon_avl: in STD_LOGIC;
      mon_we: in STD_LOGIC;
		  ddr_gnt: in std_logic;
		  ddr_read: out std_logic;
		  ddr_rqst: out std_logic;
			ddr_length: out STD_LOGIC_VECTOR (5 downto 0);
			ddr_addr: out STD_LOGIC_VECTOR (22 downto 0);
			ddr_w_data: out STD_LOGIC_VECTOR (35 downto 0);
		  ddr_we: out std_logic;
-- download data write
      debug: out STD_LOGIC_VECTOR (31 downto 0);
      w_data: in STD_LOGIC_VECTOR (31 downto 0);
			ddr_pa: in STD_LOGIC_VECTOR (9 downto 0);
      addr: in STD_LOGIC_VECTOR (15 downto 0);
      run: in STD_LOGIC;
      n_strobe: in STD_LOGIC;
      vme_cycle: in STD_LOGIC;
			mem_test: in STD_LOGIC;
      addr_str: in STD_LOGIC;
      vme_read: in STD_LOGIC;
			rd_length: in STD_LOGIC_VECTOR (3 downto 0);
      we_byte : in std_logic_vector(3 downto 0)
    );
end ddr_rw_ctrl;

architecture my_arch of ddr_rw_ctrl is
signal strobe: std_logic := '0';
signal ddr_rqst_i: std_logic := '0';
signal ddr_we_i: std_logic := '0';
signal ddr_read_i: std_logic := '0';
--signal vme_addr: std_logic := '1';
signal rst_length: std_logic := '1';
signal ddr_we_dl: std_logic := '0';
signal mem_read: std_logic := '0';
signal mem_read_sync: std_logic_vector(2 downto 0) := (others => '0');
signal end_of_block: std_logic := '0';
signal end_of_event: std_logic := '0';
signal vme_pending: std_logic := '0';
signal rd_rqst: std_logic := '0';
signal we_wqueue: std_logic := '0';
signal wqueue_din: std_logic_vector(31 downto 0) := (others => '0');
signal wqueue_dout: std_logic_vector(31 downto 0) := (others => '0');
signal wqueue_wa: std_logic_vector(8 downto 0) := (others => '0');
signal wqueue_ra: std_logic_vector(8 downto 0) := (others => '0');
signal wqueue_ra0_sync: std_logic_vector(2 downto 0) := (others => '0');
signal wqueue_wa0_sync: std_logic_vector(2 downto 0) := (others => '0');
signal wqueue_wca: std_logic_vector(2 downto 0) := (others => '0');
signal wqueue_wcb: std_logic_vector(2 downto 0) := (others => '0');
--signal wq_wa_sync: std_logic_vector(3 downto 0) := (others => '0');
--signal wq_wa_sync2: std_logic_vector(3 downto 0) := (others => '0');
signal w_length: std_logic_vector(5 downto 0) := (others => '0');
signal wqueue_dav: std_logic := '0';
signal w_addr: std_logic_vector(22 downto 0) := (others => '0');
signal r_addr: std_logic_vector(22 downto 0) := (others => '0');
signal ddr_addr_i: std_logic_vector(22 downto 0) := (others => '0');
begin
ddr_rqst <= ddr_rqst_i;
ddr_read <= ddr_read_i;
ddr_we <= ddr_we_i;
ddr_addr <= ddr_addr_i;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
		if(run = '1')then
			ddr_w_data <= x"0" & mon_data(31 downto 0);
		elsif(strobe = '1')then
			ddr_w_data <= not we_byte & w_data;
		end if;
		if(run = '1')then
			ddr_we_i <= mon_we;
		elsif(we_byte /= x"0" and (addr(13) = '1' or mem_test = '1') and strobe = '1')then
			ddr_we_i <= '1';
		else
			ddr_we_i <= '0';
		end if;
		if(strobe = '1')then
			if((mem_test = '0' and addr(5 downto 0) = "111111") or (mem_test = '1' and w_length(3 downto 0) = x"e"))then
				end_of_block <= '1';
			else
				end_of_block <= '0';
			end if;
		end if;
		if(strobe = '1' and addr_str = '1' and addr(13) = '1' and vme_read = '1' and vme_cycle = '1')then
			mem_read <= not mem_read;
		end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(reset = '1')then
      ddr_we_dl <= '0';
      we_wqueue <= '0';
      vme_pending <= '0';
		else
      ddr_we_dl <= ddr_we_i;
			if(run =	'1')then
        we_wqueue <= ddr_we_dl and not ddr_we_i;
			elsif(ddr_we_i = '1' and end_of_block = '1')then
				we_wqueue <= '1';
			else
        we_wqueue <= vme_pending and not vme_cycle and strobe;
			end if;
			if(run = '1' or mem_test = '1')then
				vme_pending <= '0';
			else
				vme_pending <= ddr_we_i or (vme_pending and not we_wqueue); 
			end if;
    end if;
		if(reset = '1')then
			wqueue_ra0_sync <= (others => '0'); 
		else
			wqueue_ra0_sync <= wqueue_ra0_sync(1 downto 0) & wqueue_ra(0); 
		end if;
		if(reset = '1')then
			wqueue_wca <= (others => '0');
		elsif(we_wqueue = '1' and wqueue_ra0_sync(2) = wqueue_ra0_sync(1))then
			wqueue_wca <= wqueue_wca + 1;
		elsif(we_wqueue = '0' and wqueue_ra0_sync(2) /= wqueue_ra0_sync(1))then
			wqueue_wca <= wqueue_wca - 1;
    end if;
		if(reset = '1')then
			wqueue_wa <= (others => '0');
		elsif(we_wqueue = '1')then
			wqueue_wa(2 downto 0) <= wqueue_wa(2 downto 0) + 1;
--			wqueue_wa(0) <= not wqueue_wa(3) xor wqueue_wa(2) xor wqueue_wa(1);
--			if(wqueue_wa(0) = '1')then
--				wqueue_wa(1) <= not wqueue_wa(3) xor wqueue_wa(2);
--			end if;
--			if(wqueue_wa(1 downto 0) = "10")then
--				wqueue_wa(2) <= not wqueue_wa(3);
--			end if;
--			if(wqueue_wa(1 downto 0) = "00")then
--				wqueue_wa(3) <= wqueue_wa(2);
--			end if;
		end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
--		if(strobe = '1')then
--			vme_addr <= not vme_cycle and not run;
--		end if;
		if(reset = '1' or we_wqueue = '1')then
			rst_length <= '1';
		elsif(ddr_we_i = '1')then
			rst_length <= '0';
		end if;
-- length is actual length minus one, reset at the first write
		if(ddr_we_i = '1')then
			if(rst_length = '1')then
				w_length <= "000000";
			else	
				w_length <= w_length + 1;
			end if;
    end if;
		if(reset = '1' or mon_we = '1')then
			mon_re <= '0';
		elsif(mon_avl = '1' and wqueue_wca(2 downto 1) /= "11")then
			mon_re <= '1';
		end if;
		if(reset = '1' or run = '0')then
			end_of_event <= '0';
		elsif(mon_we = '1' and mon_data(32) = '1')then
			end_of_event <= '1';
		elsif(we_wqueue = '1')then
			end_of_event <= '0';
		end if;
  end if;
end process;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(reset = '1' and strobe = '1')then
			w_addr <= (others => '0');				
		elsif(run = '0' and addr_str = '1'  and vme_cycle = '1' and addr(13) = '1' and strobe = '1')then
			w_addr <= ddr_pa & addr(12 downto 0); 
		elsif(we_wqueue = '1')then
			if(mem_test = '1')then
				w_addr(22 downto 4) <= w_addr(22 downto 4) + 1;
			elsif(end_of_event = '1')then
--				w_addr(22 downto 13) <= w_addr(22 downto 13) + 1;
				w_addr(19 downto 13) <= w_addr(19 downto 13) + 1;
				w_addr(12 downto 0) <= "0000000000000";
			else
				w_addr(12 downto 6) <= w_addr(12 downto 6) + 1;
				w_addr(5 downto 0) <= "000000";
			end if;
		end if;
  end if;
end process;
wqueue_din <= "000" & w_length & w_addr;
i_wqueue: ramb16_s36_s36
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => wqueue_ra,
         clka => mclk0, 
         ssra => '0', 
         ena => '1', 
         wea => '0', 
         doa => wqueue_dout, 
         dib =>  wqueue_din,
         dipb => x"0",
         addrb => wqueue_wa,
         clkb => clk,
         ssrb => '0',
         enb => we_wqueue,
         web => '1');
process(mclk0,reset)
begin
	if(reset = '1')then
		wqueue_ra <= (others => '0');
		wqueue_wa0_sync <= (others => '0');
		wqueue_wcb <= (others => '0');
--		wq_wa_sync <= x"0";
--		wq_wa_sync2 <= x"0";
		mem_read_sync <= "000";
		wqueue_dav <= '0';
		rd_rqst <= '0';
		ddr_rqst_i <= '0';
		r_addr <= (others => '0');
  elsif(mclk0'event and mclk0 = '1')then
--		if(ddr_gnt = '1')then
--			if(ddr_read_i = '0')then
--				wqueue_ra(0) <= not wqueue_ra(3) xor wqueue_ra(2) xor wqueue_ra(1);
--				if(wqueue_ra(0) = '1')then
--					wqueue_ra(1) <= not wqueue_ra(3) xor wqueue_ra(2);
--				end if;
--				if(wqueue_ra(1 downto 0) = "10")then
--					wqueue_ra(2) <= not wqueue_ra(3);
--				end if;
--				if(wqueue_ra(1 downto 0) = "00")then
--					wqueue_ra(3) <= wqueue_ra(2);
--				end if;
--			end if;
--		end if;
--		wq_wa_sync <= wqueue_wa(3 downto 0);
--		wq_wa_sync2 <= wq_wa_sync;
		if(ddr_gnt = '1' and ddr_read_i = '0')then
			wqueue_ra(2 downto 0) <= wqueue_ra(2 downto 0) + 1;
		end if;
		wqueue_wa0_sync <= wqueue_wa0_sync(1 downto 0) & wqueue_wa(0); 
		if(ddr_gnt = '1' and ddr_read_i = '0' and wqueue_wa0_sync(2) = wqueue_wa0_sync(1))then
			wqueue_wcb <= wqueue_wcb - 1;
		elsif((ddr_gnt = '0' or ddr_read_i = '1') and wqueue_wa0_sync(2) /= wqueue_wa0_sync(1))then
			wqueue_wcb <= wqueue_wcb + 1;
		end if;
		mem_read_sync <= mem_read_sync(1 downto 0) & mem_read;
		if(ddr_gnt = '1' and ddr_read_i = '0')then
			wqueue_dav <= '0';
--		elsif(wq_wa_sync2 = wqueue_ra(3 downto 0))then
		elsif(wqueue_wcb = "000")then
			wqueue_dav <= '0';
		else
			wqueue_dav <= '1';
		end if;
		if(ddr_gnt = '1' and ddr_read_i = '1')then
			rd_rqst <= '0';
		elsif(mem_read_sync(2) /= mem_read_sync(1) or (mem_test = '1' and ddr_gnt = '1' and ddr_read_i = '0'))then
			rd_rqst <= '1';
		end if;
		if(ddr_gnt = '1')then
			ddr_rqst_i <= '0';
		else
			ddr_rqst_i <= rd_rqst or wqueue_dav;
		end if;
		if(mem_test = '0')then
			r_addr <= ddr_pa & addr(12 downto 0);
		elsif(ddr_gnt = '1')then
			r_addr <= ddr_addr_i;
		end if;
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(ddr_rqst_i = '0')then
			if(rd_rqst = '1')then
				ddr_addr_i <= r_addr;
				ddr_length <= "00" & rd_length;
			else
				ddr_addr_i <= wqueue_dout(22 downto 0);
				ddr_length <= wqueue_dout(28 downto 23);
			end if;
			ddr_read_i <= rd_rqst;
		end if;
  end if;
end process;
debug <= "000000000" & w_addr;
end my_arch;
