----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    13:24:37 06/16/2009 
-- Design Name: 
-- Module Name:    ethernetnew - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.std_logic_misc.all;
use mylib.dcc_pack.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity ethernetnew is
    Port ( clk : in  STD_LOGIC;
           n_strobe : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           addr_str : in  STD_LOGIC;
           mod_SN : in  STD_LOGIC_VECTOR (7 downto 0);
           r_data : in  STD_LOGIC_VECTOR (31 downto 0);
           addr : in  STD_LOGIC_VECTOR (15 downto 0);
           e_csn : out  STD_LOGIC;
           e_wrn : out  STD_LOGIC;
           e_rdn : out  STD_LOGIC;
           e_addr : out  STD_LOGIC_VECTOR (1 downto 0);
           e_data : inout  STD_LOGIC_VECTOR (7 downto 0);
           en_write : in  STD_LOGIC;
           ethernet_sel : in  STD_LOGIC;
           ethernet_abort : in  STD_LOGIC;
           r_data_val : in  STD_LOGIC;
           ethernet_addr : out  STD_LOGIC_VECTOR (15 downto 0);
           ethernet_rd_length : out  STD_LOGIC_VECTOR (3 downto 0);
           ethernet_wdata : out  STD_LOGIC_VECTOR (31 downto 0);
           ethernet_read : out  STD_LOGIC;
           ethernet_str : out  STD_LOGIC;
           debug : out  STD_LOGIC_VECTOR (31 downto 0));
end ethernetnew;

architecture Behavioral of ethernetnew is
constant w_offset_l : positive := 32+13;
constant w_offset_h : positive := 32+14;
constant mask_DR : positive := 32+15;
constant r_base : positive := 32+16;
constant snd_data : positive := 32+17;
constant rcv_data : positive := 32+18;
constant clr_send : positive := 32+19;
constant cond : positive := 32+27;
signal strobe : std_logic := '0';
signal w5100_init : std_logic := '1';
signal et_str : std_logic_vector(2 downto 0) := (others => '0');
signal et_str_dl : std_logic := '0';
signal sm : std_logic_vector(63 downto 0) := (others => '0');
signal sm_aa : std_logic_vector(8 downto 0) := (others => '0');
signal sm_ab : std_logic_vector(8 downto 0) := (others => '0');
signal we_send_data : std_logic := '0';
signal re_send_data : std_logic := '0';
signal dav_send_datap : std_logic := '0';
signal dav_send_data : std_logic := '0';
signal send_data_out : std_logic_vector(31 downto 0) := (others => '0');
signal s_e_data_out : std_logic_vector(3 downto 0) := (others => '0');
signal byte_cntr : std_logic_vector(1 downto 0) := (others => '0');
signal aborted : std_logic := '0';
signal SN : std_logic_vector(7 downto 0) := (others => '0');
signal rec_size : std_logic_vector(2 downto 0) := (others => '0');
signal e_wrn_i : std_logic := '0';
signal ethernet_wr : std_logic := '0';
signal ethernet_rd : std_logic := '0';
signal rd_str : std_logic := '0';
signal blk_end : std_logic := '0';
signal reading : std_logic := '0';
signal e_data_in : std_logic_vector(7 downto 0) := (others => '0');
signal e_data_out : std_logic_vector(7 downto 0) := (others => '0');
signal got_buffer : std_logic := '0';
signal offset : std_logic_vector(15 downto 0) := (others => '0');
signal ethernet_addr_i : std_logic_vector(15 downto 0) := (others => '0');
signal ethernet_wdata_i : std_logic_vector(31 downto 0) := (others => '0');
signal wc_read : std_logic_vector(15 downto 0) := (others => '0');
signal last_blk : std_logic := '0';
signal start_send : std_logic := '0';
signal choice : std_logic := '0';
signal branch : std_logic := '0';
signal offset_not_0 : std_logic := '0';
signal rec_size_not_0 : std_logic := '0';
signal send_loop : std_logic := '0';
signal en_sm : std_logic := '0';
signal cmdbuf_a : std_logic_vector(10 downto 0) := (others => '0');
signal cmdbuf_aa : std_logic_vector(10 downto 0) := (others => '0');
signal we_cmdbuf : std_logic := '0';
signal rst_cmdbuf : std_logic := '0';
begin
ethernet_addr <= ethernet_addr_i;
ethernet_wdata <= ethernet_wdata_i;
e_wrn <= e_wrn_i;
e_data <= e_data_out when e_wrn_i = '0' else (others => 'Z');
process(clk)
begin
	if(clk'event and clk = '1')then
		strobe <= n_strobe;
	end if;
end process;
process(clk)
begin
	if(clk'event and clk = '1')then
		if(strobe = '1')then
			w5100_init <= w5100_init and not addr_str;
			et_str <= et_str(1 downto 0) & ((w5100_init and addr_str) or et_str_dl);
		end if;
	end if;
end process;
-- for 150MHz clock increase delay to 9
i_et_str_dl: delayN generic map(N=> 6) port map(d => et_str(2), ce => strobe, c => clk, q => et_str_dl);
i_send_data : srl_fifo
  generic map (N => 32)
  port map(
	rst => reset,
	din => r_data,
	clk => clk,
	w_strobe => strobe,
	we => we_send_data,
	r_strobe => strobe,
	re => re_send_data,
	dav => dav_send_datap,
	dout => send_data_out);
we_send_data <= r_data_val and ethernet_sel;
s_e_data_out <= aborted & '1' & byte_cntr when sm(snd_data) = '1' else aborted & '0' &  sm(29 downto 28);
process(clk)
begin
	if(clk'event and clk = '1')then
		if(strobe = '1')then
			if(et_str_dl = '1')then
				if(dav_send_data = '0')then
					byte_cntr <= "00";
				elsif(sm(snd_data) = '1')then
					byte_cntr <= byte_cntr + 1;
				end if;
			end if;
			if(byte_cntr = "11" and sm(snd_data) = '1' and et_str(2) = '1')then
				re_send_data <= '1';
			else
				re_send_data <= '0';
			end if;
			if(aborted = '0')then
				dav_send_data <= dav_send_datap;
			elsif(rd_str = '1')then
				dav_send_data <= '1';
			elsif(re_send_data = '1' and blk_end = '1')then
				dav_send_data <= '0';
			end if;
			e_addr <= sm(33 downto 32);
			if(et_str(0) = '1')then
				e_wrn_i <= '1';
				e_rdn <= '1';
			elsif(et_str(1) = '1')then
				e_wrn_i <= not sm(34);
				e_rdn <= not sm(35);
			end if;
			if(et_str(2) = '1')then
				e_csn <= not sm(35) and not sm(34);
			elsif(et_str_dl = '1')then
				e_csn <= '1';
			end if;
			if(reading = '0')then
				aborted <= '0';
			elsif(ethernet_abort = '1')then
				aborted <= '1';
			end if;
			e_data_in <= e_data;
			if(sm(w_offset_h) = '1' and et_str_dl = '1')then
				offset(15 downto 8) <= "01" & sm(r_base) & e_data_in(4 downto 0);
			elsif(sm(w_offset_l) = '1' and et_str_dl = '1')then
				offset(7 downto 0) <= e_data_in;
			elsif((sm(snd_data) = '1' or sm(rcv_data) = '1') and et_str(1) = '1')then
				offset(12 downto 0) <= offset(12 downto 0)+1;
			end if;
			if(sm(w_offset_h) = '1' and et_str(2) = '1' and sm(r_base) = '0')then
				got_buffer <= '1';
			elsif(sm(clr_send) = '1' and et_str(2) = '1')then
				got_buffer <= '0';
			end if;
			if(et_str(1) = '1')then
				case s_e_data_out is
					when x"0" => e_data_out <= sm(23 downto 16);
					when x"1" => e_data_out <= SN;
					when x"2" => e_data_out <= offset(15 downto 8);
					when x"3" => e_data_out <= offset(7 downto 0);
					when x"4" => e_data_out <= send_data_out(31 downto 24);
					when x"5" => e_data_out <= send_data_out(23 downto 16);
					when x"6" => e_data_out <= send_data_out(15 downto 8);
					when x"7" => e_data_out <= send_data_out(7 downto 0);
					when x"8" => e_data_out <= sm(23 downto 16);
					when x"9" => e_data_out <= SN;
					when x"a" => e_data_out <= offset(15 downto 8);
					when x"b" => e_data_out <= offset(7 downto 0);
					when others => e_data_out <= x"ff";
				end case;
			end if;
			if(sm(rcv_data) = '0')then
				rec_size <= "111";
			elsif(et_str(0) = '1')then
				rec_size <= rec_size - 1;
			end if;
			if(sm(rcv_data) = '1' and et_str_dl = '1')then
				ethernet_addr_i <= ethernet_addr_i(7 downto 0) & ethernet_wdata_i(31 downto 24);
				ethernet_wdata_i <= ethernet_wdata_i(23 downto 0) & e_data_in;
				if(ethernet_addr_i = x"a000" and rec_size = "001" and en_write = '1')then
					ethernet_wr <= '1';
				else
					ethernet_wr <= '0';
				end if;
				if(ethernet_addr_i = x"0500" and rec_size = "001")then
					ethernet_rd <= '1';
				else
					ethernet_rd <= '0';
				end if;
				ethernet_read <= ethernet_rd;
			elsif(re_send_data = '1' and blk_end = '1')then
				ethernet_addr_i(14 downto 4) <= ethernet_addr_i(14 downto 4) + 1;
			end if;
			if(ethernet_rd = '1' and et_str(0) = '1')then
				reading <= '1';
			elsif(re_send_data = '1' and blk_end = '1' and wc_read(15 downto 11) = ethernet_wdata_i(31 downto 27) and last_blk = '1')then
				reading <= '0';
			end if;
			if(reading = '0')then
				wc_read <= (others => '0');
			elsif(re_send_data = '1')then
				wc_read <= wc_read + 1;
			end if;
			if((last_blk = '1' and wc_read(3 downto 0) = ethernet_wdata_i(19 downto 16)) or wc_read(3 downto 0) = x"f")then
				blk_end <= '1';
			else
				blk_end <= '0';
			end if;
			if((sm(w_offset_h) = '1' and et_str(2) = '1' and sm(r_base) = '0') or (re_send_data = '1' and blk_end = '1' and last_blk = '0'))then
				rd_str <= '1';
			else
				rd_str <= '0';
			end if;
			if(re_send_data = '1' and blk_end = '1' and last_blk = '1')then
				start_send <= '1';
			elsif(sm(clr_send) = '1' and et_str_dl = '1')then
				start_send <= '0';
			end if;
			if(wc_read(15 downto 4) = ethernet_wdata_i(31 downto 20) or wc_read(10 downto 4) = "1111111")then
				last_blk <= '1';
			else
				last_blk <= '0';
			end if;
			if(wc_read(15 downto 4) = ethernet_wdata_i(31 downto 20))then
				ethernet_rd_length <= ethernet_wdata_i(19 downto 16);
			else
				ethernet_rd_length <= x"f";
			end if;
			if(aborted = '0' and ((ethernet_wr = '1' and et_str_dl = '1') or rd_str = '1'))then
				ethernet_str <= '1';
			else
				ethernet_str <= '0';
			end if;
			if(sm(cond) = '1' and ((e_data_in = sm(23 downto 16) and sm(mask_DR) = '0') or (sm(mask_DR) = '1' and or_reduce(e_data_in and sm(23 downto 16)) = '0')))then
				choice <= '1';
			else
				choice <= branch;
			end if;
			if(or_reduce(offset(12 downto 0)) = '1')then
				offset_not_0 <= '1';
			else
				offset_not_0 <= '0';
			end if;
			if(offset_not_0 = '1' and dav_send_data = '1')then
				send_loop <= '1';
			else
				send_loop <= '0';
			end if;
			if(rec_size = "000")then
				rec_size_not_0 <= '0';
			else
				rec_size_not_0 <= '1';
			end if;
			case sm(26 downto 24) is
				when "001" => branch <= rec_size_not_0;
				when "010" => branch <= dav_send_data;
				when "011" => branch <= send_loop;
				when "100" => branch <= not start_send;
				when "101" => branch <= reading;
				when "110" => branch <= got_buffer;
				when "111" => branch <= offset_not_0;
				when others => branch <= '0';
			end case;
		end if;
		en_sm <= et_str(0) and n_strobe;
	end if;
end process;
sm_aa <= sm(15 downto 8) & '0' when choice = '1' else sm(7 downto 0) & '0';
sm_ab <= sm(15 downto 8) & '1' when choice = '1' else sm(7 downto 0) & '1';
SN <= mod_sn when mod_sn(7) = '1' else x"7" & mod_sn(3 downto 0);
i_state_machine: RAMB16_S36_S36
   generic map (
			INIT_00 => x"0000000500000004000000040003000308000008000002010000000400800001",
			INIT_01 => x"00000007000100080000000700A800070000000700C000060000000600010005",
			INIT_02 => x"0000000700FF000C0000000700FF000B0000000700FF000A0000000700010009",
			INIT_03 => x"00000007004800100000000700DE000F0000000700AC000E000000070000000D",
			INIT_04 => x"0000000700C00014000000070000001300000007100000120000000700000011",
			INIT_05 => x"0000000600160018000000071000001700000007000100160000000700A80015",
			INIT_06 => x"000000070008001C0000000700D0001B000000070007001A0000000700000019",
			INIT_07 => x"0000000600000020000000050004001F000000070003001E000000070003001D",
			INIT_08 => x"0000000700170024000000070000002300000006000400220000000700010021",
			INIT_09 => x"0800000B00002827000000070001002700000006000100260000000400010025",
			INIT_0A => x"000000070002002C000000060001002B0800000B00132A5F0000000600030029",
			INIT_0B => x"0800000B001732300800000B0016312F000000060003002E0800000B00002D2C",
			INIT_0C => x"000000060026003400000004000300330800000B001732310800000B0014315F",
			INIT_0D => x"0001400B2000003800000006002800370800800B00F844360800000B00003536",
			INIT_0E => x"0004000B01003B3C000000063000003B000000052000003A0000200B30000039",
			INIT_0F => x"0000000730000040000000072000003F000000060028003E000000050004003D",
			INIT_10 => x"0800000B00004442000000060001004300000007004000420000000600010041",
			INIT_11 => x"0800800B00F85A4800000006002000470000000006004B46000000000500455A",
			INIT_12 => x"0000000002004C5A0000200B3000004B0000400B2000004A0000000600240049",
			INIT_13 => x"00000000070052500002000703004E4F000000063000004E000000052000004D",
			INIT_14 => x"00000006002400540000000504045A530000000602004E520000000500400051",
			INIT_15 => x"0008000700200058000000060001005700000007300000560000000720000055",
			INIT_16 => x"0800000B001C5F5C000000060003005B0800000B00005A580000000600010059",
			INIT_17 => x"000000060001006000000000050044320800000B0000245E000000060003005D",
			INIT_18 => x"00000000000000000800000B0000246100000006000100620000000700100061")
 port map (
      DOA => sm(31 downto 0),      -- Port A 32-bit Data Output
      DOB => sm(63 downto 32),      -- Port B 32-bit Data Output
      ADDRA => sm_aa,  -- Port A 9-bit Address Input
      ADDRB => sm_ab,  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => x"00000000",      -- Port A 32-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => en_sm,      -- Port A RAM Enable Input
      ENB => en_sm,      -- PortB RAM Enable Input
      SSRA => w5100_init,    -- Port A Synchronous Set/Reset Input
      SSRB => w5100_init,    -- Port B Synchronous Set/Reset Input
      WEA => '0',      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );
process(clk)
begin
	if(clk'event and clk = '1')then
		if(strobe = '1')then
			if(reset = '1')then
				cmdbuf_a <= (others => '0');
			elsif(sm(rcv_data) = '1' and et_str_dl = '1' and cmdbuf_a(10) = '0')then
				cmdbuf_a <= cmdbuf_a + 1;
			end if;
		end if;
	end if;
end process;
we_cmdbuf <= sm(rcv_data) and et_str_dl;
rst_cmdbuf <= '0' when addr(14 downto 10) = "10101" else '1';
cmdbuf_aa <= cmdbuf_a(10 downto 2) & not cmdbuf_a(1 downto 0);
i_cmdbuf: RAMB16_S9_S36
   port map (
      DOB => debug,      -- Port B 32-bit Data Output
      ADDRA => cmdbuf_aa,  -- Port A 11-bit Address Input
      ADDRB => addr(8 downto 0),  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => e_data_in,      -- Port A 8-bit Data Input
      DIB => x"00000000",      -- Port B 32-bit Data Input
      DIPA => "0",    -- Port A 1-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => strobe,      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => rst_cmdbuf,    -- Port B Synchronous Set/Reset Input
      WEA => we_cmdbuf,      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );

end Behavioral;

