-- due to tRAS minimum 40ns, active to prechage must
-- have at least eight clock cycles(200MHz). This also guarantees active to active is
-- eleven clock cycles minimum 
-- read length is limited to 16 words in this module
-- when the time period is too long without issuing memory read, a fake read for timing calibration should be issued 
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;
use mylib.DCC_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity sdram_b is
		generic (sim : boolean := false; clk_speed : positive := 1 ; clk_period : real := 10.0);
    port (
-- reset input, should be connected to the inverted LOCK output of dll
			reset: in STD_LOGIC;
-- system clock input
			clk: in STD_LOGIC;
			mclk0: in STD_LOGIC;
			mclk90: in STD_LOGIC;
			mclk180: in STD_LOGIC;
			mclk270: in STD_LOGIC;
			n_strobe: in STD_LOGIC;
-- DLL lock input
			locked: in STD_LOGIC;
		  mem_status : out STD_LOGIC_VECTOR (15 downto 0);
-- VME read signals
		  rst_rfifo_data: in std_logic; -- force rfifo_data to all '0' when not selected
		  rfifo_data : out STD_LOGIC_VECTOR (31 downto 0);
			rfifo_dav: out STD_LOGIC;
-- memory page register page size = 16KBytes
-- write queue signals
		  ddr_gnt: out std_logic;
		  ddr_w_we: in std_logic;
		  ddr_read: in std_logic;
		  ddr_rqst: in std_logic;
			ddr_length: in STD_LOGIC_VECTOR (5 downto 0);
			ddr_addr: in STD_LOGIC_VECTOR (22 downto 0);
			ddr_w_data: in STD_LOGIC_VECTOR (35 downto 0);
-- SDRAM ports
--			tc : out STD_LOGIC; -- 7.8 us strobe
			mclkio : inout STD_LOGIC;
			mdq : inout STD_LOGIC_VECTOR (15 downto 0);
			ma : out STD_LOGIC_VECTOR (12 downto 0);
			mba : out STD_LOGIC_VECTOR (1 downto 0);
			dm : out STD_LOGIC_VECTOR (1 downto 0);
			dqs : inout STD_LOGIC_VECTOR (1 downto 0);
			we_n : out STD_LOGIC;
			ras_n : out STD_LOGIC;
			cas_n : out STD_LOGIC;
			cs_n : out STD_LOGIC;
			mclke : out STD_LOGIC;
			mclk : out STD_LOGIC;
			mclk_n : out STD_LOGIC
    );
end sdram_b;

architecture sdram_arch of sdram_b is
signal locked_dl: std_logic := '0';
signal rst_rclka: std_logic := '0';
signal r_locked_dl: std_logic := '0';
signal rst_rclkb: std_logic := '0';
signal rclk0_dcm: std_logic := '0';
signal rclk90_dcm: std_logic := '0';
signal rclk270_dcm: std_logic := '0';
signal rclk0: std_logic := '0';
signal rclk90: std_logic := '0';
signal rclk270: std_logic := '0';
signal mclkin: std_logic := '0';
signal mclkout: std_logic := '0';
-- sdram i/o
signal strobe : std_logic := '0';
signal ma_i: std_logic_vector(14 downto 0) := (others => '0');
signal md: std_logic_vector(31 downto 0) := (others => '0');
signal byte_n: std_logic_vector(3 downto 0) := (others => '0');
signal mq: std_logic_vector(31 downto 0) := (others => '0');
signal mq_q: std_logic_vector(31 downto 0) := (others => '0');
signal rfifo_din: std_logic_vector(31 downto 0) := (others => '0');
signal rasn: std_logic := '1';
signal casn: std_logic := '1';
signal wen: std_logic := '1';
signal set_dm: std_logic := '0';
signal active_dl: std_logic_vector(3 downto 0) := (others => '0');
-- refresh signal generation
signal ref: std_logic := '0';
signal ref_toggle: std_logic := '0';
signal ref_tgl_sync: std_logic_vector(2 downto 0) := (others => '0');
signal start_init: std_logic := '0';
-- sdram state machine
signal en_dll: std_logic := '0';
signal rst_dll: std_logic := '0';
signal w_cycle: std_logic := '0';
signal r_cycle: std_logic := '0';
signal wr_burst: std_logic := '0';
signal stop_w: std_logic := '0';
signal stop_r: std_logic := '0';
signal we_rfifo: std_logic := '0';
signal ec_dqs_sync : std_logic_vector(2 downto 0) := (others => '0');
signal active: std_logic := '0';
signal inh_ras: std_logic := '0';
signal ref_rqst: std_logic := '0';
signal rd_burst : std_logic_vector(2 downto 0) := (others => '0');
signal init_done: std_logic_vector(1 downto 0) := (others => '1');
signal pc_dl: std_logic := '0';
signal precharge: std_logic := '0';
signal precharge_dl: std_logic_vector(1 downto 0) := (others => '0');
signal rfc_cntr: std_logic_vector(4 downto 0) := (others => '0');
signal tRAS: std_logic := '0';
signal wfifo_do: std_logic_vector(35 downto 0) := (others => '0');
signal wfifo_do_q: std_logic_vector(35 downto 0) := (others => '0');
signal wfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_wa_sync : std_logic_vector(3 downto 0) := (others => '0');
signal rfifo_wa_sync2 : std_logic_vector(3 downto 0) := (others => '0');
signal rfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal burst_cnt : std_logic_vector(6 downto 0) := (others => '0');
--signal rd_burst_cnt : std_logic_vector(4 downto 0) := (others => '0');
signal burst_length : std_logic_vector(5 downto 0) := (others => '0');
signal read_flag: std_logic := '0';
signal wr_burst_dl: std_logic := '0';
signal dcm_status : std_logic_vector(7 downto 0) := (others => '0');
signal tune_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal ec_dqs: std_logic := '0';
signal ec_dqs_p: std_logic_vector(1 downto 0) := (others => '0');
signal chk_ps: std_logic := '0';
signal r_locked: std_logic := '0';
signal dqs_in : std_logic_vector(3 downto 0) := (others => '0');
signal dqs_status : std_logic_vector(1 downto 0) := (others => '0');
signal rclk_tune: std_logic_vector(1 downto 0) := (others => '0');
signal psincdec: std_logic := '0';
signal psen_tog_sync: std_logic_vector(3 downto 0) := (others => '0');
signal psen_toggle: std_logic := '0';
signal psen: std_logic := '0';
signal ps_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal ps_delay : std_logic_vector(7 downto 0) := (others => '0');
signal t_md: std_logic := '0';
signal dqs_o: std_logic_vector(1 downto 0) := (others => '0');
signal t_dqs: std_logic_vector(1 downto 0) := (others => '0');
signal dqs_outp: std_logic := '0';
signal dqs_out: std_logic := '0';
signal mem_rdy: std_logic := '0';
signal mem_rdy_q: std_logic := '0';
attribute LOC : string;
--attribute LOC of i_DCM_rclkb : label is "DCM_x0y2"; -- left top
attribute LOC of i_DCM_rclka : label is "DCM_x1y3"; -- top right
attribute LOC of i_rclk90 : label is "BUFGMUX_x2y10"; -- top F
--attribute LOC of i_rclk270 : label is "BUFGMUX_x0y2"; -- left A
attribute LOC of i_rclk0 : label is "BUFGMUX_x1y10"; -- top H
begin
i_mclkout:  oddr2
    port map(
          d0 => '1',
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclkout);
i_mclkio : IOBUF
   port map (
      O => mclkin,     -- Buffer output
      IO => mclkio,   -- Buffer inout port (connect directly to top-level port)
      I => mclkout,     -- Buffer input
      T => '0'      -- 3-state enable input 
   );
----------------------------------------------------------------------------------------
i_rst_rclka : SRL16
   port map (
      Q => locked_dl,       -- SRL data output
      A0 => '1',     -- Select[0] input
      A1 => '1',     -- Select[1] input
      A2 => '1',     -- Select[2] input
      A3 => '1',     -- Select[3] input
      CLK => mclk0,   -- Clock input
      D => locked        -- SRL data input
   );
rst_rclka <= not locked_dl;
i_DCM_rclka: dcm 
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",CLKIN_PERIOD => clk_period,
					PHASE_SHIFT => 0,DESKEW_ADJUST => "SOURCE_SYNCHRONOUS")  
	 port map	
         ( CLKIN   => mclkin , -- always use mclk0 as CLKIN, to move the rclk phase, change the rclk assignments to mdq and dqs instead
           CLKFB   => rclk0_dcm ,
           RST     => rst_rclka,
					 STATUS => dcm_status, -- 8-bit DCM status bits output
           DSSEN   => '0' ,
           PSINCDEC=> psincdec ,
           PSEN    => psen ,
           PSCLK   => clk ,
					 clk0    => rclk0_dcm,
					 clk90    => rclk90_dcm,
					 LOCKED  => r_locked
					 );
i_rclk0 : BUFG port map (O=>rclk0, I=>rclk0_dcm);
--i_rst_rclkb : SRL16
--   port map (
--      Q => r_locked_dl,       -- SRL data output
--      A0 => '1',     -- Select[0] input
--      A1 => '1',     -- Select[1] input
--      A2 => '1',     -- Select[2] input
--      A3 => '1',     -- Select[3] input
--      CLK => rclk0,   -- Clock input
--      D => r_locked(0)        -- SRL data input
--   );
--rst_rclkb <= not r_locked_dl;
--i_DCM_rclkb: dcm 
--   generic map(CLKOUT_PHASE_SHIFT => "FIXED",DESKEW_ADJUST => "SOURCE_SYNCHRONOUS",
--					CLKIN_PERIOD => 10.0,PHASE_SHIFT => 47)  
--	 port map	
--         ( CLKIN   => rclk0 , -- always use mclk0 as CLKIN, to move the rclk phase, change the rclk assignments to mdq and dqs instead
--           CLKFB   => rclk90 ,
--           RST     => rst_rclkb,
--           DSSEN   => '0' ,
--           PSINCDEC=> psincdec ,
--           PSEN    => psen ,
--           PSCLK   => clk ,
--					 clk0    => rclk90_dcm,
--           CLK180   => rclk270_dcm,
--					 LOCKED	 => r_locked(1)
--					 );
i_rclk90 : BUFG port map (O=>rclk90, I=>rclk90_dcm);
--i_rclk270 : BUFG port map (O=>rclk270, I=>rclk270_dcm);
rclk270 <= not rclk90;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
    if(strobe = '1')then
      mem_status <= mem_rdy_q & init_done(1) & dqs_status & '0' & dcm_status(2 downto 0) & ps_cntr;
		end if;
		psen_tog_sync <= psen_tog_sync(2 downto 0) & psen_toggle;
		psen <= psen_tog_sync(3) xor psen_tog_sync(2);
    mem_rdy_q <= mem_rdy;
	end if;
end process;
process(clk,locked)
begin
	if(locked = '0')then
		ps_cntr <= x"80";
  elsif(clk'event and clk = '1')then
		if(psen = '1')then
			if(psincdec = '1')then
				ps_cntr <= ps_cntr + 1;
			else
				ps_cntr <= ps_cntr - 1;
			end if;
		end if;
	end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(ec_dqs_sync(2) = '1')then
			tune_cntr <= (others => '0');
		elsif(ref = '1')then
			tune_cntr <= tune_cntr + 1;
		end if;
		if(init_done(1) = '0' or rclk_tune(1) = '1')then
			rclk_tune(0) <= '0';
		elsif(rfc_cntr(4) = '1')then -- set to '1' during refresh cycle
			rclk_tune(0) <= not mem_rdy or tune_cntr(7);
		end if;
		if(active = '1')then
			rclk_tune(1) <= rclk_tune(0);
		elsif(ec_dqs_sync(2 downto 1) = "10")then
			rclk_tune(1) <= '0';
		end if;
		ec_dqs_sync <= ec_dqs_sync(1 downto 0) & ec_dqs_p(1);
  end if;
end process;
i_wfifo : RAMB16_S36_S36
   generic map(SIM_COLLISION_CHECK => "NONE")
   port map (
      DIA => ddr_w_data(31 downto 0),      -- Port A 32-bit Data Input
      DIPA => ddr_w_data(35 downto 32),    -- Port-A 4-bit parity Input
      ADDRA => wfifo_wa,  -- Port A 10-bit Address Input
      CLKA => clk,    -- Port A Clock
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      ENA => ddr_w_we,      -- Port A RAM Enable Input
      WEA => '1',      -- Port A Write Enable Input
      DIB => x"00000000",      -- Port B 16-bit Data Input
      DIPB => x"0",    -- Port B 2-bit parity Input
      ADDRB => wfifo_ra,  -- Port B 10-bit Address Input
      CLKB => mclk0,    -- Port B Clock
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      ENB => '1',      -- PortB RAM Enable Input
      WEB => '0',    -- Port B Write Enable Input
      DOB => wfifo_do(31 downto 0),      -- Port B 32-bit Data Output
      DOPB => wfifo_do(35 downto 32)      -- Port B 4-bit Data Output
   );
process(clk)
begin
  if(clk'event and clk = '1')then
		if(reset = '1')then
			wfifo_wa <= (others => '0');
		elsif(ddr_w_we = '1')then
			wfifo_wa <= wfifo_wa + 1;
		end if;
  end if;
end process;
process(mclk0,reset)
begin
	if(reset = '1')then
		wfifo_ra <= (others => '0');
		burst_cnt <= (others => '0');
  elsif(mclk0'event and mclk0 = '1')then
		if(burst_cnt(6) = '1' and w_cycle = '1')then
			wfifo_ra <= wfifo_ra + 1;
		end if;	
		if(active_dl(0) = '1')then	 
			burst_cnt <= '1' & not burst_length(5 downto 0);
		elsif(burst_cnt(6) = '1')then
      burst_cnt <= burst_cnt + 1;	 
    end if;
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		wfifo_do_q <= wfifo_do;
		md(31 downto 16) <= wfifo_do_q(31 downto 16);
		byte_n(3 downto 2) <= wfifo_do_q(35 downto 34);
		ref_toggle <= ref_toggle xor ref;
  end if;
end process;
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
		md(15 downto 0) <= wfifo_do_q(15 downto 0);
		byte_n(1 downto 0) <= wfifo_do_q(33 downto 32);
  end if;
end process;
process(rclk90,reset)
begin
  if(reset = '1')then
    rfifo_wa <= (others => '0');
  elsif(rclk90'event and rclk90 = '1')then
		we_rfifo <= ec_dqs and not rclk_tune(1);
    if(we_rfifo = '1')then
			rfifo_wa(0) <= not rfifo_wa(3) xor rfifo_wa(2) xor rfifo_wa(1);
			if(rfifo_wa(0) = '1')then
				rfifo_wa(1) <= not rfifo_wa(3) xor rfifo_wa(2);
			end if;
			if(rfifo_wa(1 downto 0) = "10")then
				rfifo_wa(2) <= not rfifo_wa(3);
			end if;
			if(rfifo_wa(1 downto 0) = "00")then
				rfifo_wa(3) <= rfifo_wa(2);
			end if;
    end if;
  end if;
end process;
i_rd_burst_dl: delayN
  generic map(N=> 5) port map(d => rd_burst(2), ce => '1', c => rclk90, q => ec_dqs);
process(rclk90)
begin
  if(rclk90'event and rclk90 = '1')then
		mq_q(15 downto 0) <= mq(15 downto 0);
		rfifo_din <= mq_q;
		rd_burst(2) <= rd_burst(1) and init_done(1);
  end if;
end process;
process(mclk270)
begin
  if(mclk270'event and mclk270 = '1')then
		rd_burst(1) <= rd_burst(0);
  end if;
end process;
process(rclk270)
begin
  if(rclk270'event and rclk270 = '1')then
		mq_q(31 downto 16) <= mq(31 downto 16);
  end if;
end process;
i_rfifo: ramb16_s36_s36
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => rfifo_ra,
         clka => clk, 
         ssra => rst_rfifo_data, 
         ena => strobe, 
         wea => '0', 
         doa => rfifo_data, 
         dib =>  rfifo_din,
         dipb => x"0",
         addrb => rfifo_wa,
         clkb => rclk90,
         ssrb => '0',
         enb => we_rfifo,
         web => '1');
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
			rfifo_wa_sync <= rfifo_wa(3 downto 0);
			rfifo_wa_sync2 <= rfifo_wa_sync;
			if(rst_rfifo_data = '1')then
				rfifo_ra <= "00000" & rfifo_wa_sync2;
				rfifo_dav <= '0';
			elsif(rfifo_ra(3 downto 0) /= rfifo_wa_sync2)then
				rfifo_ra(0) <= not rfifo_ra(3) xor rfifo_ra(2) xor rfifo_ra(1);
				if(rfifo_ra(0) = '1')then
					rfifo_ra(1) <= not rfifo_ra(3) xor rfifo_ra(2);
				end if;
				if(rfifo_ra(1 downto 0) = "10")then
					rfifo_ra(2) <= not rfifo_ra(3);
				end if;
				if(rfifo_ra(1 downto 0) = "00")then
					rfifo_ra(3) <= rfifo_ra(2);
				end if;
				rfifo_dav <= '1';
			else
				rfifo_dav <= '0';
			end if;
    end if;
  end if;
end process;
process(rclk90,r_locked)
begin
  if(r_locked = '0')then
		chk_ps <= '0';
		dqs_status <= "00";
		ref_tgl_sync <= "000";
		psen_toggle <= '0';
  elsif(rclk90'event and rclk90 = '1')then
		chk_ps <= ref_tgl_sync(1) xor ref_tgl_sync(2);
		if(chk_ps = '1')then
			dqs_status <= "00";
		elsif(ec_dqs = '1')then
			dqs_status(0) <= dqs_status(0) or not dqs_in(3);
			dqs_status(1) <= dqs_status(1) or dqs_in(3);
		end if;
		ref_tgl_sync <= ref_tgl_sync(1 downto 0) & ref_toggle;
		if(chk_ps = '1')then
			case dqs_status is
				when "01" =>
					psincdec <= '1';
					psen_toggle <= dcm_status(0) xnor psen_toggle;
				when "10" =>
					psincdec <= '0';
					psen_toggle <= dcm_status(0) xnor psen_toggle;
				when others => null;	 
			end case;
		end if;
	end if;
end process;
mem_rdy <= ps_delay(3) when sim else ps_delay(7);
process(rclk270)
begin
  if(rclk270'event and rclk270 = '1')then
		dqs_in(2) <= dqs_in(1) and init_done(1);
  end if;
end process;
process(rclk90)
begin
  if(rclk90'event and rclk90 = '1')then
    dqs_in(3) <= dqs_in(2);
		ec_dqs_p(0) <= ec_dqs;
		ec_dqs_p(1) <= ec_dqs_p(0) or ec_dqs;
  end if;
end process;
process(rclk0)
begin
  if(rclk0'event and rclk0 = '1')then
		dqs_in(1) <= dqs_in(0);
  end if;
end process;
-- memory address
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
-- this ensures active, read/write and precharge are to the same bank
    if(init_done(1) = '0')then
			ma_i(14 downto 13) <= '0' & en_dll; 
    elsif(inh_ras = '0')then
      ma_i(14 downto 13) <= ddr_addr(22 downto 21);
		end if;
    if(init_done(1) = '0')then
-- latency = 3 burst = 8
			ma_i(12 downto 0) <= "00" & precharge & '0' & rst_dll & "00" & rst_dll & rst_dll & "00" & (rst_dll or en_dll) & rst_dll; 
		else
			if(inh_ras = '0')then
				ma_i(12 downto 0) <= ddr_addr(20 downto 8);
			else
				if(active_dl(1) = '1')then
					ma_i(8 downto 0) <= ddr_addr(7 downto 0) & '0';
				elsif(burst_cnt(6) = '1')then 	
					ma_i(6 downto 1) <= ma_i(6 downto 1) + 1;
				end if;
				ma_i(10) <= precharge;				
			end if;
		end if;
  end if;
end process;
-- generate read and write burst signals -------------------------------------------------------
-- burst_cntr
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(rclk_tune(0) = '1')then
			burst_length <= "000111";
			read_flag <= '1';
		elsif(active = '1')then		
			if(ddr_read = '1')then
				burst_length <= "00" & ddr_length(3 downto 0);
			else
				burst_length <= ddr_length;
			end if;
			read_flag <= ddr_read;
    end if;
		active_dl <= active_dl(2 downto 0) & active;
		wr_burst <= burst_cnt(6) and w_cycle;
		rd_burst(0) <= burst_cnt(6) and r_cycle;
		wr_burst_dl <= wr_burst;
		ddr_gnt <= active_dl(3) and not rclk_tune(1);
		set_dm <= not wr_burst_dl;
  end if;
end process;
process(mclk270,reset)
begin
	if(reset = '1')then
		t_md <= '1';
		dqs_outp <= '0';
		dqs_out <= '0';
  elsif(mclk270'event and mclk270 = '1')then
		t_md <= (not active_dl(2) or not w_cycle) and (t_md or precharge_dl(0));
		dqs_outp <= (not casn and w_cycle) or (dqs_outp and not precharge);
		dqs_out <= dqs_outp;
  end if;
end process;
-- sdram state machine -----------------------------------------------------------------------------------------
-- sdram power on initialization states: precharge all banks, enable DLL, reset DLL, precharge all banks,
-- two refresh cycle and set mode register.
i_en_dll: delayN
  generic map(N=> 5) port map(d => start_init, ce => '1', c => mclk0, q => en_dll);
i_rst_dll: delayN
  generic map(N=> 7) port map(d => start_init, ce => '1', c => mclk0, q => rst_dll);
process(mclk0,locked,rasn,casn,wen)
variable state: std_logic_vector(2 downto 0);
begin
state := rasn & casn & wen;
  if(locked = '0')then
    ps_delay <= x"00";
    w_cycle <= '0';
    r_cycle <= '0';
    pc_dl <= '0';
    precharge <= '0';
    precharge_dl <= (others => '0');
    rfc_cntr <= (others => '0');
    tRAS <= '0';
		inh_ras <= '0';
    active <= '0';
    stop_w <= '0';
    stop_r <= '0';
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    init_done <= "00";
    ref_rqst <= '0';
  elsif(mclk0'event and mclk0 = '1')then
		if(mem_rdy = '0' and stop_r = '1')then
			ps_delay <= ps_delay + 1;
		end if;
		pc_dl <= start_init or rst_dll or stop_w or stop_r;
    precharge <= pc_dl;
    precharge_dl <= precharge_dl(0) & precharge;
-- auto refresh tRFC counter
    if(state = "001")then
			rfc_cntr <= "10011";
		elsif(rfc_cntr(4) = '1')then
			rfc_cntr <= rfc_cntr + 1;
		end if;
-- inhibit RAS after refresh or read/write until tRFC or tPR
    if(rasn = '0' and wen = '1')then
			inh_ras <= '1';
    elsif(precharge_dl(1) = '1' or rfc_cntr(3 downto 1) = "111")then
			inh_ras <= '0';
		end if;
    if(active = '1')then
			tRAS <= '1';
    elsif(active_dl(3) = '1')then
			tRAS <= '0';
		end if;
-- latch refresh request
    if(ref = '1')then
      ref_rqst <= '1';
    end if;
		active <= '0';
		if(r_cycle = '1' and stop_r = '0' and (burst_cnt(6) = '0' or and_reduce(burst_cnt(5 downto 1)) = '1') and tRAS = '0')then
			stop_r <= '1';
		else
			stop_r <= '0';
		end if;
		stop_w <= wr_burst_dl and not wr_burst;
    if(active_dl(0) = '1')then
			r_cycle <= read_flag;
		elsif(stop_r = '1')then
			r_cycle <= '0';
		end if;
    if(active_dl(0) = '1')then
			w_cycle <= not read_flag;
		elsif(stop_w = '1')then
			w_cycle <= '0';
		end if;
    case state is
-- NOP
      when "111" =>
---- goto precharge
        if(precharge = '1')then
          rasn <= '0';
          wen <= '0';
        end if;
---- goto auto refresh
        if(ref_rqst = '1' and inh_ras = '0')then
          rasn <= '0';
          casn <= '0';
        end if;
---- goto set mode
        if(en_dll = '1' or rst_dll = '1')then
          rasn <= '0';
          casn <= '0';
          wen <= '0';
        end if;
---- goto active
        if((ddr_rqst = '1' or rclk_tune(0) = '1') and ref_rqst = '0' and inh_ras = '0')then
          rasn <= '0';
					active <= '1';
        end if;
------ goto write/read
        if(active_dl(1) = '1' or (burst_cnt(6) = '1' and ma_i(2 downto 1) = "11"))then
          casn <= '0';
          wen <= r_cycle;
        end if;
-- AUTO REFRESH
      when "001" =>
				ref_rqst <= '0';
        init_done <= init_done(0) & '1';
				rasn <= '1';
				casn <= '1';
      when others =>
        if(burst_cnt(6) = '0' or ma_i(2 downto 1) /= "11")then
-- default to go to NOP
					rasn <= '1';
					casn <= '1';
					wen <= '1';
        end if;
    end case;
  end if;
end process;
----------------------------------------------------------------------------------------
-- ram pads
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    we_n <= wen;
    cas_n <= casn;
    ras_n <= rasn;
    ma <= ma_i(12 downto 0);
		mba <= ma_i(14 downto 13);
  end if;
end process;
g_dm: for i in 0 to 1 generate
  i_dm: oddr2
    port map(
			c0 => mclk0,
			c1 => mclk180,
			q => dm(i),
			ce => '1',
			d0 => byte_n(2+i),
			d1 => byte_n(i),
			r => '0',
			s => set_dm);
end generate;
g_mdq_l: for i in 0 to 7 generate
  ddr_data_l: ddr_io
    port map(
          clk => mclk0,
          clkn => mclk180,
          pad => mdq(i),
					t => t_md,
          d0 => md(i+16),
          d1 => md(i),
          iclk => rclk90,
          iclkn => rclk270,
          q0 => mq(i),
          q1 => mq(16+i));
end generate;
g_mdq_h: for i in 8 to 15 generate
  ddr_data_h: ddr_io
    port map(
          clk => mclk0,
          clkn => mclk180,
          pad => mdq(i),
					t => t_md,
          d0 => md(i+16),
          d1 => md(i),
          iclk => rclk90,
          iclkn => rclk270,
          q0 => mq(i),
          q1 => mq(16+i));
end generate;
i_dqs0: oddr2
    port map(
			c0 => mclk270,
			c1 => mclk90,
			q => dqs_o(0),
			ce => '1',
			d0 => dqs_out,
			d1 => '0',
			r => '0',
			s => '0');
i_dqs0_buf : IOBUF
   port map (
      O => dqs_in(0),     -- Buffer output
      IO => dqs(0),   -- Buffer inout port (connect directly to top-level port)
      I => dqs_o(0),     -- Buffer input
      T => t_dqs(0)      -- 3-state enable input 
   );
process(mclk270)
begin
  if(mclk270'event and mclk270 = '1')then
		t_dqs <= t_md & t_md;
  end if;
end process;
i_dqs1: oddr2
    port map(
			c0 => mclk270,
			c1 => mclk90,
			q => dqs_o(1),
			ce => '1',
			d0 => dqs_out,
			d1 => '0',
			r => '0',
			s => '0');
i_dqs1_buf : OBUFT
   port map (
      O => dqs(1),   -- Buffer out port (connect directly to top-level port)
      I => dqs_o(1),     -- Buffer input
      T => t_dqs(1)      -- 3-state enable input 
   );
i_mclk:  oddr2
    port map(
          d0 => '1',
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk);
i_mclk_n:  oddr2
    port map(
          d0 => '0',
          d1 => '1',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mclk_n);
-- instantiate refresh generator
I_refresh: refresh
	generic map(sim => sim, N => clk_speed)
	PORT MAP(
		clk => mclk0,
		lock => locked,
		init_done => init_done(1),
		start_init => start_init,
		ddr_csn => cs_n,
		ddr_clke => mclke,
--		tc_out => tc,
		ref => ref
	);
end sdram_arch;
